<?php

require_once 'phonetizer.php';

define('GLOBAL_MUL', 0.7); // 0.7
define('SQ_MUL', 0.8);
define('TR_MUL', 0.8);

function log2p12($h)
{
  return 12*log($h)/log(2);
}

class MelodicCh extends Phonetizer
{
  private $current;
  public $stream;
  
  public $pos;
  private $reset;
  
  function __construct()
  {
    $this->stream = Array();
    $this->reset  = Array('on'=>false, 'off'=>false, 'hz'=>Array(),
                           'lasthz' => 0,
                           'lastvol' => 16,
                           'attack' => 16,
                           'type' => 0);
    $this->current = $this->reset;
    $this->pos = 0;
  }
  
  function SetData($timer, $vol, $hz, $type)
  {
    $cur = &$this->current;
    
    #print "$vol, last={$cur['lastvol']}, att={$cur['attack']}, on={$cur['on']}\n";
    
    if(($cur['on']!==false
                     && $vol <= $cur['lastvol'] - 1
                     && $vol <= $cur['attack'] - 9
                  )
    || !$vol)
    {
      if($cur['on'] !== false)
      {
        $cur['hz'][(string)$timer] = $cur['lasthz'];
        $cur['off'] = $timer;
        $this->stream[ (string)$cur['on'] ] = $cur;
      }
      $cur = $this->reset;
    }
    else
    {
      if($cur['on'] !== false
         && ($vol >= $cur['lastvol']+4
           || abs(log2p12($hz) - log2p12($cur['lasthz'])) >= 0.9
            )
         )
      {
        $cur['hz'][(string)$timer] = $cur['lasthz'];
        $cur['off'] = $timer;
        $this->stream[ (string)$cur['on'] ] = $cur;
        $cur = $this->reset;
      }
      
      if($cur['on']===false)
      {
        $cur['on'] = $timer;
        $cur['attack'] = $vol;
      }
      if($hz != $cur['lasthz'])
      {
        $cur['hz'][(string)$timer] = $hz;
        $cur['lasthz'] = $hz;
      }
      $cur['lastvol'] = $vol;
      $cur['type'] = $type;
    }
  }
  
  function Fin($timer)
  {
    $this->SetData($timer, 0,0, 0);
    $prevpos = false;
    foreach($this->stream as $pos => &$data)
    {
      if($prevpos !== false)
        $this->stream[$prevpos]['next'] = $data['on'];
      $prevpos = $pos;
    }
    $this->stream[$prevpos]['next'] = $timer + 100;
  }
  
};

class SquareCh extends MelodicCh
{
  private $identity;
  
  function __construct($id)
  {
    $this->identity = $id;
    MelodicCh::__construct();
  }
  
  function Phonetize()
  {
    #print_r($this->stream);
    
    $result = '';
    foreach($this->stream as &$item)
    {
      $begin = (float)$item['on'];
      $end   = (float)$item['off'];
      
      if($begin > $this->pos+0.001)
      {
        $result .= $this->CreateSilencePho($begin - $this->pos);
      }
      if($begin < $this->pos) $begin = $this->pos;
      
      $duration = $end - $begin;
      
      $duration = max(
        $duration,
        /*max(0.06, */ min(0.3, ((float)$item['next']) - $begin)//)
         );
      //$duration = max($duration, 0.1);
      
      $hzlist = Array();
      foreach($item['hz'] as $pos => $hz)
      {
        $pos = (float)$pos;
        $pos -= $begin;
        if($pos < 0.0001) $pos = 0;
        if($pos >= $end-0.001) $pos = $duration;
        $hzlist[(string)$pos] = $hz*SQ_MUL;
      }
      $hzlist[(string)$duration] = $hz*SQ_MUL;
      
      //$absbegin = (($begin*1000 - 2400) + 4800) / 4800;//- mario 
      //$absbegin = 1;
      
if($this->identity == 0) /* jp2 */
{
      if(false && $this->identity == 0)
      {
        static $counter=0;
        $set = Array('so:', 'ro', 'mo', 'n'); //'ko','ni','tSi','wa');
        $c = $set[$counter++%4];
        $vow = ereg_replace('[^aiueo:]', '', $c);
        $con = str_replace($vow, '', $c);
        //if($con == $vow) $vow = '';
      }
      else
      {
        $vowels = Array('a:', 'e:', 'o:', 'u', 'i');
        $cons   = Array('d'); //k', 'p', 't', 'n', 't', 'm', 'r', 'k', 'b', 'h', 'n', 'd', 'j'/*, 's', 'S', 'ts', 'tS'*/);
        srand((int)($begin*5)*7);
        do {
          $vow = $vowels[rand(0,count($vowels)-1)];
          $con = $cons[rand(0,count($cons)-1)];
        } while(ereg('hu|du|ni|hi|je|tse|si|ji|dzi', $con.$vow));
      }
}
if($this->identity == 1) /* jp1 or jp3 */
{
      $vowels = Array('e'); //, 'e', 'o', 'u', 'i');
      $cons   = Array('s'); //'m', 'w', 't', 'k'); //'n', 't', 'm', 'r', 'k', 'b', 'h', 'n', 'd', 'j'/*, 's', 'S', 'ts', 'tS'*/);
      srand((int)($begin*5)*7);
      do {
        $vow = $vowels[rand(0,count($vowels)-1)];
        $con = $cons[rand(0,count($cons)-1)];
      } while(ereg('wu|we|wo', $con.$vow));
}
      
      if(!empty($con) && $duration >= 0.08)
        $seq = Array('0.0'   => $con,
                     '0.075' => $vow);
      else
        $seq = Array('0.0'   => $vow);
      
      if(empty($vow)) unset($seq['0.075']);
      
      if($con=='k') $seq['0.2'] = 'N';
      
      $result .= $this->CreatePhoSeq($seq, $duration, $hzlist);
    }
    return $result;
  }
};

class TriangleCh extends MelodicCh
{
  function Phonetize()
  {
    #print_r($this->stream);
    
    $result = '';
    foreach($this->stream as &$item)
    {
      $begin = (float)$item['on'];
      $end   = (float)$item['off'];
      
      if($begin > $this->pos)
      {
        $result .= $this->CreateSilencePho($begin - $this->pos);
      }
      if($begin < $this->pos) $begin = $this->pos;
      
      $duration = $end - $begin;
      $duration = max(
        $duration,
        max(0.08, min(3.2, ((float)$item['next']) - $begin))
         );
      
      $hzlist = Array();
      foreach($item['hz'] as $pos => $hz)
      {
        $pos = (float)$pos;
        $pos -= $begin;
        if($pos >= $end-0.001) $pos = $duration;
        $hzlist[(string)$pos] = $hz*TR_MUL;
      }
      $hzlist[(string)$duration] = $hz*TR_MUL;
      
      $result .= $this->CreatePhoSeq(
        Array('0.0'   => 'p',
              '0.035' => 'a',
              '0.15'  => 'N'),
        $duration, $hzlist);
    }
    return $result;
  }
};

class NoiseCh extends MelodicCh
{
  function Phonetize()
  {
    #print_r($this->stream);
    
    $result = '';
    foreach($this->stream as $item)
    {
      $begin = (float)$item['on'];
      $end   = (float)$item['off'];
      
      if($begin > $this->pos)
      {
        $result .= $this->CreateSilencePho($begin - $this->pos);
      }
      if($begin < $this->pos) $begin = $this->pos;
      
      $duration = $end - $begin;
      $duration = max(
        $duration,
        max(0.08,  min(0.15, ((float)$item['next']) - $begin))
         );
      if($duration > 0.7) $duration = 0.7;
      
      if($item['lasthz'] < 3000)
        $result .= $this->CreatePhoSeq(
          Array('0.00'   => 'tS',
                '0.07'   => 'i',
               //  '0.2'    => 'N:'
              ),
          $duration, Array(0 => 120, (string)$duration => 120));
      else
        $result .= $this->CreatePhoSeq(
          Array('0.0'   => 't',
                '0.075' => 'a',
               // '0.1'   => 'N:'
                ),
          $duration, Array(0 => 40, (string)$duration => 40));
    }
    return $result;
  }
};

class Song
{
  private $rate;
  private $timer;
  
  private $sq0;
  private $sq1;
  private $tri;
  private $noi;
  private $pcm;
  
  public function __construct()
  {
    $this->rate = 60;
    $this->timer= 0;
    
    $this->sq0 = new SquareCh(0);
    $this->sq1 = new SquareCh(1);
    $this->tri = new TriangleCh;
    $this->noi = new NoiseCh;
    $this->pcm = null;
  }
  
  function Fin()
  {
    $this->sq0->Fin($this->timer);
    $this->sq1->Fin($this->timer);
    $this->tri->Fin($this->timer);
    $this->noi->Fin($this->timer);
    
    file_put_contents('n.pho', $this->noi->Phonetize());
    file_put_contents('t.pho', $this->tri->Phonetize());
    file_put_contents('a.pho', $this->sq0->Phonetize());
    file_put_contents('b.pho', $this->sq1->Phonetize());
  }
  
  public function SetRate($rate)
  {
    $this->rate = $rate;
  }
  
  public function AddAge($n)
  {
    $this->timer += $n / $this->rate;
  }
  
  public function SetData($chn, $duty, $vol, $wavelen)
  {
    #if($chn != 0) return;

    #if($chn < 2) $wavelen *= 2;
    
    if(!$wavelen)$wavelen=1;
    $hz = 110000 / $wavelen;
    
    $hz *= GLOBAL_MUL;
    
    #if($chn == 0) $hz *= 2;
    
    #if($hz >= 500) $hz = 500;
    
    #print "[{$this->timer}] $chn: $duty $vol $hz\n";
    
    switch($chn)
    {
      case 0:
        $this->sq0->SetData($this->timer, $vol, $hz, $duty);
        break;
      case 1:
        $this->sq1->SetData($this->timer, $vol, $hz, $duty);
        break;
      case 2:
        $this->tri->SetData($this->timer, $vol, $hz, $duty);
        break;
      case 3:
        $this->noi->SetData($this->timer, $vol, $hz, $duty);
        break;
    }
  }
};

function ReadSND($string)
{
  $song = new Song();
  
  extract(unpack('A4sig/Cvers/Cnchans/Crate', substr($string, 0, 16)));
  if($sig != 'SND') die('Not a SND file');
  
  $song->SetRate($rate);
  
  $pos = 16;
  $end = strlen($string);
  while($pos < $end)
  {
    $c = ord($string[$pos++]);
    if($c == 0xFB)
    {
      extract(unpack('Cn/vlen', substr($string,$pos,3)));
      $pos += 3 + $len;
      continue;
    }
    if($c == 0xFE)
    {
      $pos += 2; // skip channel type
      continue;
    }
    if($c == 0xFF)
    {
      $song->AddAge(1);
      continue;
    }
    if($c == 0xFD)
    {
      $song->AddAge(ord($string[$pos++]));
      continue;
    }
    if($vers == 3 &&
       (($c & 0xF0) == 0x20)
    || (($c & 0xF0) == 0x40)
    || (($c & 0xF0) == 0x80)
    || (($c & 0xF0) == 0xC0))
    {
      $duty = $c >> 4;
      $vol  = $c & 15;
      extract(unpack('Cchn/vwavelen', substr($string,$pos,3))); $pos += 3;
      $song->SetData($chn, $duty, $vol, $wavelen);
      continue;
    }
    if($c < $nchans)
    {
      extract(unpack('vwavelen/Cvol', substr($string,$pos,3))); $pos += 3;
      if($vers == 1) { $vol /= 16; $wavelen = 110000 / $wavelen; }
      
      if($vers == 1 && $c > 1) $wavelen /= 2;

      if($c == 2) $wavelen *= 2;
      
      if(!$wavelen)$wavelen=1;
      $song->SetData($c, 4, $vol, $wavelen);
      continue;
    }
  }
  $song->Fin();
}

#$fn = 'mm2dw1.snd';
#$fn = 'ddg2.snd';
#$fn = 'gl10.snd';
#$fn = 'solo.snd';
#$fn = 'smb3g.snd';
#$fn = 'smb3t.snd';
#$fn = 'mm2dw1.snd';
#$fn = 'mm4pr.snd';
#$fn = 'kirby1.snd';
#$fn = 'kirby2.snd';
#$fn = 'kirby3.snd';
#$fn = 'balloon.snd';
$fn = 'castleva.snd';

ReadSND(file_get_contents($fn));
